# This code calculates the maximum mean downstream dispersal per
# generation (Ladv) of larvae that is compatible with a population
# being retained, as described in Pringle, Lutscher and Glick.  It
# plots the critical value of Ladv assuming that the kernel is
# Gaussian, the value given by equation (2) of Pringle, Lutscher and
# Glick, and the exact value calculated numerically.  To use it, you
# must first define the dispersal kernel, as described in the comments
# below.
#
# For this code to run, you must have the numpy, scipy, and matplotlib
# python modules installed.  These can be obtained as part of
# Enthought Python on Windows machines, installed as part of fink for
# OS X machines, and are included in most linux distributions.

import numpy
from pylab import *
import scipy.optimize as sciop
import scipy.integrate as integrate
import scipy.interpolate as si

# choose which integration routine to use to integrate the kernel.
# Unless your kernel has an odd shape, stick to the default.  To learn
# more, read the documentation for scipy.integrate
integrator=integrate.quad

# first, one must define a dispersal kernel as the probability
# distribution function K(x) which returns a single value, the density
# of larvae attempting to recruit at location x.  The integral of K(x)
# over all space must be one, and the mean location of larval
# dispersal must be 0 (i.e. the integral of x*K(x) over all space must
# be 0).  Thus K(x) is the dispersal kernel in the absance of larval
# transport.
#
# the following is a Gaussian kernel for testing.  All methods of
# calculationg the speed should give the same result for a Gaussian
# Kernel
def kern_gauss(x):
    #define a Gaussian dispersal kernel with a standard deviation of
    #recruitment distance of Ld
    Ld=30.0;
    K=(1/Ld/sqrt(2*pi))*exp(-(x**2)/2.0/Ld**2)
    return K

# The following defines an arbitrary kernel, as one would if one had
# observed a dispersal kernel or had calculated one in a numerical
# model.  It is important that the integral of the kernel be 1, and
# that the kernel be centered arround x=0.  If this is not done, an
# error message will be printed by the code.  This error message will
# include the value of the integral of the kernel, and its mean.
# These can then be used to adjust the kernel defined below.
#
# x_kern defines the points at which the kernel is defined.  Outside
# of these points the kernel is assumed to be zero.
#
# K_kern is the density at the points defined in x_kern.  These values
# are linearly integrated at points between where they are defined.
x_kern=array([0.0, 5.0, 10.0, 15.0, 20.0, 25.0])-12.3333333266
K_kern=array([0.0, 1.0,  7.0,  6.0,  1.0,  0.0])/75.0
kern_interp=si.interp1d(x_kern,K_kern,kind='linear',bounds_error=False,
                        fill_value=0.0)

#CHOOSE WHICH KERNEL TO USE!
kern=kern_gauss

def spd_calc(N,kern):
    """
    This code takes values N and a kernel shape and returns the
    critical value of La that just allows retention.
    
    The inputs are:
    
    N; N for the critter, as defined in Pringle, Lutscher and Glick
    
    kern; the dispersal kernel, a function of distance
    
    It returns
    La_crit ; the critical value of La that allow retention
    Ld_emp; the standard deviation of the kernel
    gamma2; the excess kurtosis of the kernel    
    """

    #Define Ld_multiple.  Some integrations will be restricted to with
    #+/-Ld_multiple*(the standard deviation of the dispersal distance)
    Ld_multiple=10.0
    
    Ksum,jnk=integrator(kern,-integrate.inf,integrate.inf)
    def meanfunc(x,kern):
        return x*kern(x);
    Kmean,jnk=integrator(meanfunc,-integrate.inf,integrate.inf,(kern,))

    def stdfunc(x,kern):
        return x**2*kern(x)
    mu2,jnk=integrator(stdfunc,-integrate.inf,integrate.inf,(kern,))
    Kstd=sqrt(mu2)

    def fourthmofunc(x,kern):
        return x**4*kern(x)
    mu4,jnk=integrator(fourthmofunc,-integrate.inf,integrate.inf,(kern,))
    Kgamma2=mu4/Kstd**4-3.0

    #check for errors in the dispersal kernel
    #sum of dispersal probability is 1?
    if abs(Ksum-1.0)>0.0001:
        print ' '
        print 'The kernel does not integrate to 1'
        print 'The kernel integrates to',Ksum
        print 'NONE OF THE FOLLOWING RESULTS WILL BE VALID!'
        print ' '
    #kernel is centered arround 0?
    if abs(Kmean/Kstd)>0.001:
        print ' '
        print 'The mean distance of larval dispersal in the kernel'
        print 'Is not zero, as it should be!'
        print 'The mean dispersal distance is',Kmean
        print 'NONE OF THE FOLLOWING RESULTS WILL BE VALID!'
        print ' '
    


    #define moment generating function, calculated over a domain that
    #stretches from -Kstd*Ld_multiple to Kstd*Ld_multiple
    def Moment_gen_true(s,Kstd,kern):
        def Mgen_integrand(x,s,kern):
            return numpy.exp(s*x)*kern(x)
        Ld_multiple=30;
        (M,jnk)=integrator(Mgen_integrand,
                               -Ld_multiple*Kstd,Ld_multiple*Kstd,
                               (s,kern))
        return M


    #find minimum of (1/s)*log(N*Moment_gen) in s in range from 0
    #to max_search/Ld (?)
    max_search=20.0
    def to_optimize_true(s,N,Kstd,kern):
        return (1/s)*numpy.log(N*Moment_gen_true(s,Kstd,kern))
    smin_true=sciop.fminbound(to_optimize_true,0.0,max_search/Kstd,
                              (N,Kstd,kern))

    if (abs(smin_true-max_search/Kstd)/(max_search/Kstd))<0.05:
        print 'smin_true',smin_true,'search limit',max_search/Kstd
        print 'smin_true is close to the search limit'
        print 'perhaps you should expand the search?'

    La_crit_true=to_optimize_true(smin_true,N,Kstd,kern)

    return La_crit_true,Kstd,Kgamma2

#Now, define the values of N for which you would like to find the
#critical values of Ladv, both as defined in Pringle, Lutscher and
#Glick.
N_vec=numpy.logspace(0,1,num=30)

#these are the arrays to store the results in
La_theory_gauss=0.0*N_vec+numpy.nan #Ladv_critical assuming a Gaussian Kernel
La_theory=0.0*N_vec+numpy.nan #Ladv_critical using the result in Pringle et al.
La_calc=0.0*N_vec+numpy.nan #the exact value of Ladv computed numerically

for n in range(N_vec.size):
    N=N_vec[n]

    La_calc[n],Ld_calc,gamma2_calc=spd_calc(N,kern);
    print 'For N=',N,' La,Ld,gamma2 are',La_calc[n],Ld_calc,gamma2_calc
    print ' '

    La_theory_gauss[n]=sqrt(log(N)*2.0*Ld_calc**2)
    La_theory[n]=sqrt(2.0*Ld_calc**2*(log(N)
                                 +(log(N)**2)*gamma2_calc/6.0
                                 +(log(N)**3)*(gamma2_calc**2)/144.0))

    #plot results as we go along.
    clf()
    l1=plot(La_theory,N_vec,'-',linewidth=2)
    l2=plot(La_theory_gauss,N_vec,'--',linewidth=2)
    l3=plot(La_calc,N_vec,'b+',markersize=15,linewidth=2)
    if True:
        legend((l3,l1,l2),('Full Solution',
                           'Kurtosis approx. from (2)',
                           'Gaussian approx. from BWP'),
               loc='upper left')
    title('Critical N for given Ladv',fontsize='large')
    ylabel('Critical N',fontsize='large')
    xlabel('Mean Larval Dispersal Distance Ladv',fontsize='large')
    grid(b=True)
    draw()
    
show()




